from __future__ import annotations

import contextlib
from functools import lru_cache
from typing import Any, Optional, TYPE_CHECKING, Iterable

from pyrus.models.requests import FormRegisterRequest, TaskCommentRequest

if TYPE_CHECKING:
    from pyrus import PyrusAPI
    from pyrus.models.entities import FormRegisterFilter


class PyrusORMSession:
    def __init__(self, pyrus_api: PyrusAPI):
        self.pyrus_api = pyrus_api

    @lru_cache(maxsize=512)
    def get_catalog(self, catalog_id: int):
        from pyrus_orm.catalog import CatalogItem

        catalog = self.pyrus_api.get_catalog(catalog_id)
        headers = [x.name for x in catalog.catalog_headers]
        return [
            CatalogItem(
                item_id=item.item_id,
                catalog_id=catalog_id,
                headers=headers,
                values_row=item.values,
                values=dict(zip(headers, item.values))
            )
            for item in catalog.items
        ]

    def get_task_raw(self, task_id: int) -> dict[str, Any]:
        response = self.pyrus_api._perform_get_request(
            self.pyrus_api._create_url(f'/tasks/{task_id}')
        )
        if 'error' in response:
            raise Exception(response['error'])  # TODO: proper error handling

        if not response.get('task'):
            raise Exception('no task received')

        return response['task']

    def update_task(self, task_id: int, field_updates: list[Any], comment: Optional[str] = None) -> None:
        response = self.pyrus_api._perform_post_request(
            self.pyrus_api._create_url(f'/tasks/{task_id}/comments'),
            {
                'text': comment,
                'field_updates': field_updates,
            }
        )
        if 'error' in response:
            raise Exception(response['error'])  # TODO: proper error handling

        if not response.get('task'):
            raise Exception('no task received')

        return response['task']

    def create_task(self, data: dict[str, Any]) -> dict[str, Any]:
        response = self.pyrus_api._perform_post_request(
            self.pyrus_api._create_url(f'/tasks'),
            data
        )
        if 'error' in response:
            raise Exception(response['error'])  # TODO: proper error handling

        if not response.get('task'):
            raise Exception('no task received')

        return response['task']

    def get_filtered_tasks(
        self,
        form_id: int,
        include_archived: bool = False,
        steps: Iterable[int] = (),
        filters: Iterable[FormRegisterFilter] = (),
    ):
        request = FormRegisterRequest(
            include_archived=include_archived,
            steps=list(steps),
            filters=filters,
        )
        response = self.pyrus_api._perform_post_request(
            self.pyrus_api._create_url(f'/forms/{form_id}/register'),
            request,
        )
        if 'error' in response:
            raise Exception(response['error'])  # TODO: proper error handling

        return response.get('tasks', [])

    def comment_task(self, task_id: int, comment: str) -> None:
        self.pyrus_api.comment_task(task_id, TaskCommentRequest(text=comment))


_session: Optional[PyrusORMSession] = None


def get_session() -> PyrusORMSession:
    global _session
    assert _session is not None, 'no pyrus-orm session is set. ' \
                                 'try calling set_session_global or use `with set_session(..):`'
    return _session


@contextlib.contextmanager
def set_session(session: PyrusORMSession):
    assert isinstance(session, PyrusORMSession)
    global _session
    prev_value = _session

    _session = session
    yield
    _session = prev_value


def set_session_global(session: PyrusORMSession) -> None:
    global _session
    _session = session



from typing import TypeVar, Type, Generic, Optional, Iterable

from pyrus.models.entities import EqualsFilter

from .catalog import CatalogItem
from .session import get_session

T = TypeVar('T', bound='PyrusModel')


class Manager(Generic[T]):
    _model: Type[T]

    def __init__(self, model: Type[T]):
        self._model = model

    def get(self, task_id: int) -> Optional[T]:
        from pyrus_orm.session import get_session

        data = get_session().get_task_raw(task_id)
        return self._model.from_pyrus_data(data)

    def get_filtered(
        self,
        *,
        include_archived: bool = False,
        steps: Iterable[int] = (),
        **kwargs,
    ) -> list[T]:
        fields = self._model.Meta.fields

        filters = []
        for k, v in kwargs.items():
            assert k in fields, f'field {k} not found in model fields'

            if isinstance(v, CatalogItem):
                value = v.item_id
            else:
                value = v

            filters.append(EqualsFilter(fields[k].id, value))

        tasks = get_session().get_filtered_tasks(
            self._model.Meta.form_id,
            include_archived=include_archived,
            steps=steps,
            filters=filters
        )
        return [self._model.from_pyrus_data(x) for x in tasks]


class _ManagerProperty(Generic[T]):
    def __init__(self, model: Type[T]):
        self._model = model

    def __get__(self, instance, owner) -> Manager[T]:
        return Manager(self._model)



from dataclasses import dataclass, field
from typing import Optional, TypedDict, Any, TYPE_CHECKING, Callable

if TYPE_CHECKING:
    pass


class PyrusCatalogSingleItem(TypedDict):
    item_id: int
    item_ids: list[int]
    headers: list[str]
    values: list[str]
    rows: list[list[str]]


class _CatalogListWrapper(list['CatalogItem']):
    def find(self, pattern: dict[str, Any]) -> Optional['CatalogItem']:
        for item in self:
            for k, v in pattern.items():
                if item.values[k] != v:
                    break
            else:
                return item
        return None


class _CatalogMixin:
    catalog_id: Optional[int]

    def __init__(self, catalog_id: int):
        self.catalog_id = catalog_id

    def catalog(self) -> _CatalogListWrapper:
        from pyrus_orm.session import get_session
        assert self.catalog_id, 'catalog_id is not set'

        return _CatalogListWrapper(get_session().get_catalog(self.catalog_id))


@dataclass
class CatalogItem(_CatalogMixin):
    item_id: int
    catalog_id: Optional[int] = field(compare=False, default=None)
    headers: list[str] = field(compare=False, default_factory=list)
    values_row: list[str] = field(compare=False, default_factory=list)
    values: dict[str, str] = field(compare=False, default_factory=dict)

    _bound_field_setter: Optional[Callable[['CatalogItem'], None]] = field(compare=False, default=None)

    @classmethod
    def from_pyrus_data(
        cls,
        data: PyrusCatalogSingleItem,
        catalog_id: Optional[int],
        bound_field_setter: Optional[Callable[['CatalogItem'], None]]
    ) -> 'CatalogItem':
        return cls(
            catalog_id=catalog_id,
            item_id=data['item_id'],
            headers=data.get('headers', []),
            values_row=data.get('values', []),
            values=(
                dict(zip(data['headers'], data['values']))
                if 'headers' in data and 'values' in data else {}
            ),
            _bound_field_setter=bound_field_setter,
        )

    def find_and_set(self, pattern: dict[str, Any]):
        assert self._bound_field_setter, f"CatalogItem is not bound to model's field " \
                                         f"(you should access it as model property only)"
        item = self.catalog().find(pattern)
        if item:
            self._bound_field_setter(item)
        else:
            raise ValueError("no item matching pattern found")


class CatalogEmptyValue(_CatalogMixin):
    catalog_id: int

    def __bool__(self):
        return False



import copy
from datetime import datetime
from typing import Any, TypeVar, Type, Optional, Generic

from pyrus_orm.catalog import CatalogItem, CatalogEmptyValue
from pyrus_orm.fields import BaseField
from pyrus_orm.manager import Manager
from pyrus_orm.session import get_session
from pyrus_orm.utils import classproperty

T = TypeVar('T', bound='PyrusModel')


class PyrusModel(Generic[T]):
    id: Optional[int]
    create_date: datetime
    last_modified_date: datetime

    _data = None
    _field_values: dict[int, Any] = {}
    _changed_fields: set[int]

    class Meta:
        form_id: int
        fields: dict[str, BaseField]

    @classproperty
    def objects(cls: Type[T]) -> Manager[T]:
        return Manager(cls)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        try:
            int(cls.Meta.form_id)
        except (ValueError, AttributeError) as e:
            raise Exception('Model.Meta.form_id is not set') from e

    def __init__(self, **kwargs):
        self._field_values = {}
        self._changed_fields = set()
        self.id = None

        for field_name, field in self.Meta.fields.items():
            if field_name not in self._field_values:
                self._field_values[field.id] = {
                    'id': field.id,
                }

        for k, v in kwargs.items():
            if v is not None:
                setattr(self, k, v)

    @classmethod
    def from_pyrus_data(
        cls: Type[T],
        data: dict[str, Any],
    ) -> T:
        data = copy.deepcopy(data)

        def fix_datetime(v: str) -> str:
            # makes datetime compatible with python's 3.9 fromisoformat() function
            # 3.10+ does not require this
            return v.replace('Z', '+00:00')

        create_date = datetime.fromisoformat(fix_datetime(data['create_date']))
        last_modified_date = datetime.fromisoformat(fix_datetime(data['last_modified_date']))

        fields = data.pop('fields')

        obj = cls(
            id=data['id'],
            _data=data,
            _field_values={
                f['id']: f
                for f in fields
            },
            create_date=create_date,
            last_modified_date=last_modified_date,
        )
        return obj

    def as_pyrus_data(self):
        return {
            **(self._data or {}),
            'fields': self.get_pyrus_fields_data(),
            'form_id': self.Meta.form_id,
        }

    def get_pyrus_fields_data(self, changed_only: bool = False) -> list[Any]:
        values = list(self._field_values.values())
        if changed_only:
            values = [x for x in values if x['id'] in self._changed_fields]

        return sorted(values, key=lambda f: f['id'])

    def as_dict(self) -> dict[str, Any]:
        values = {}
        for field_name in self.Meta.fields.keys():
            value = getattr(self, field_name)
            if isinstance(value, CatalogItem):
                value = value.values
            elif isinstance(value, CatalogEmptyValue):
                value = None
            values[field_name] = value

        return values

    def save(self, comment: Optional[str] = None):
        if self.id:
            get_session().update_task(self.id, self.get_pyrus_fields_data(changed_only=True), comment)
        else:
            data = get_session().create_task(self.as_pyrus_data())
            new_item = type(self).from_pyrus_data(data)
            self.__dict__ = new_item.__dict__

    def comment(self, comment: str) -> None:
        assert self.id
        get_session().comment_task(self.id, comment)


    def get_url(self) -> str:
        return f'https://pyrus.com/t#id{self.id}'

    def __repr__(self) -> str:
        return f'Task {self.get_url()}'



from enum import Enum
from typing import Literal, Optional, Generic, TypeVar, Union, TYPE_CHECKING, Type

from .catalog import CatalogItem, CatalogEmptyValue, _CatalogListWrapper
from .session import get_session
from .types import Flag, Status

if TYPE_CHECKING:
    from .model import PyrusModel

FieldType = Literal[
    'text', 'money', 'number', 'date', 'time',
    'checkmark', 'due_date', 'due_date_time',
    'email', 'phone', 'flag', 'step', 'status',
    'creation_date', 'note', 'catalog',
]

T = TypeVar('T')


class BaseField(Generic[T]):
    id: int
    name: Optional[str] = None
    type: FieldType

    def __init__(self, id: int):
        self.id = id

    def __set_name__(self, owner, name):
        if not hasattr(owner.Meta, 'fields'):
            setattr(owner.Meta, 'fields', {})
        owner.Meta.fields[name] = self
        self.name = name

    def __get__(self, instance: 'PyrusModel', owner) -> Optional[T]:
        return instance._field_values[self.id].get('value')

    def __set__(self, instance, value: T):
        instance._field_values[self.id]['value'] = value
        instance._changed_fields.add(self.id)


class TextField(BaseField[str]):
    type = 'text'


class NumericField(BaseField[float]):
    type = 'number'


class StepField(BaseField[int]):
    type = 'step'

    def __set__(self, *_):
        raise AttributeError('step field is read-only')


class StatusField(BaseField[Status]):
    type = 'status'

    def __set__(self, *_):
        raise AttributeError('status field is read-only')


class NoteField(BaseField[str]):
    type = 'note'

    def __set__(self, *_):
        raise AttributeError('note field is read-only')


class FlagField(BaseField[Flag]):
    type = 'flag'


class CatalogField(BaseField):
    type = 'catalog'
    _catalog_id: int

    def __init__(self, id: int, *, catalog_id: int):
        super().__init__(id)
        self._catalog_id = catalog_id

    def __get__(self, instance: 'PyrusModel', owner) -> Union[CatalogEmptyValue, CatalogItem]:
        try:
            field_value = instance._field_values[self.id]['value']
        except (KeyError, AttributeError):
            return CatalogEmptyValue(self._catalog_id)

        return CatalogItem.from_pyrus_data(
            catalog_id=self._catalog_id,
            data=field_value,
            bound_field_setter=lambda value: self.__set__(instance, value),
        )

    def __set__(self, instance: 'PyrusModel', value: Union[CatalogItem, int]):
        assert isinstance(value, (int, CatalogItem)), type(value)

        if isinstance(value, CatalogItem):
            item_id = value.item_id
        else:
            item_id = value

        # TODO: check if item_id is valid
        instance._field_values[self.id]['value'] = {
            'item_id': item_id
        }
        instance._changed_fields.add(self.id)


T_Enum = TypeVar('T_Enum', bound=Enum)


class CatalogEnumField(BaseField[T_Enum]):
    type = 'catalog'
    _catalog_id: int
    _enum: Type[T_Enum]
    _id_field: str

    def __init__(self, id: int, *, catalog_id: int, enum: Type[T_Enum], id_field: str):
        super().__init__(id)
        self._catalog_id = catalog_id
        self._enum = enum
        self._id_field = id_field

    def __get__(self, instance: 'PyrusModel', owner) -> Optional[T_Enum]:
        try:
            field_value = instance._field_values[self.id]['value']
        except (KeyError, AttributeError):
            return None

        if self._id_field == 'item_id':
            enum_value = field_value['item_id']
        else:
            item_values = dict(zip(field_value['headers'], field_value['values']))
            enum_value = item_values[self._id_field]

        for item in self._enum:
            if item.value == enum_value:
                return item
        raise ValueError(f"can't find enum item with value '{enum_value}'")

    def __set__(self, instance: 'PyrusModel', value: T_Enum):
        if self._id_field == 'item_id':
            item_id = value.value
        else:
            catalog = _CatalogListWrapper(get_session().get_catalog(self._catalog_id))
            item = catalog.find({self._id_field: value.value})
            if item is None:
                raise ValueError(f"can't find catalog item with field '{self._id_field}'='{value.value}'")
            item_id = item.item_id

        instance._field_values[self.id]['value'] = {
            'item_id': item_id
        }
        instance._changed_fields.add(self.id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrus_orm']

package_data = \
{'': ['*']}

install_requires = \
['pyrus-api>=2.13.0,<3.0.0']

setup_kwargs = {
    'name': 'pyrus-orm',
    'version': '0.0.3',
    'description': "Radically simple ORM for Pyrus's tasks",
    'long_description': 'pyrus-orm\n=========\n\nRadically simple, django/peewee-like, easy and incomplete ORM for [Pyrus](https://pyrus.com).\n\nWith pyrus-orm, you can read, create and modify [tasks](https://pyrus.com/en/help/api/models#form-registry-task).\n\nWorks with [pyrus-api](https://github.com/simplygoodsoftware/pyrusapi-python) under the hood.\n\n### This is an early development version\n\n### Features:\n\n- Define models with:\n    - [x] simple fields (text, number, dates, ...)\n    - [x] catalog fields, single item\n    - [ ] catalog fields, multiple items\n    - [ ] "title" fields\n- Operations with models:\n    - [x] Create and save\n    - [x] Read from registry by ID\n    - [x] Modify and save changes\n- Filtering:\n    - [x] by include_archived and steps fields\n    - [x] by value of simple or catalog fields\n    - [ ] less than, greater than\n    - [ ] value in a list\n    - [ ] ranges\n\nInstallation\n-----------\n\n```shell\npip install pyrus-orm\n```\n\nExamples\n-------\n\n\n### Define model and initialize\n\n```python\n\nclass Book(PyrusModel):\n    title = TextField(1)  # 1 is a field ID in pyrus\'s form\n    date = DateField(2)\n    author = CatalogField(3, catalog=<catalog id>)\n\n    class Meta:\n        form_id = <form_id>\n\n\npyrus_api = PyrusAPI(...)\nsession = PyrusORMSession(pyrus_api)\n\nset_session_global(session)\n```\n\n\n### Create item\n\n```python\nbook = Book(\n    title=\'Don Quixote\',\n    date=\'1605-01-01\',\n    author=Book.author.find({\'Name\': \'Alonso Fernández de Avellaneda\'})\n)\n\nbook.save()\n\nbook.id\n>>> <task_id>\n```\n\n\n### Read and modify item\n\n```python\nbook = Book.objects.get(id=...)\n\n# simple field\nbook.title\n>>> \'Don Quixote\'\nbook.title = \'Don Quixote, Part Two\'\nbook.save(\'title changed\')\n\n# catalog field\nbook.author\n>>> CatalogItem(item_id=..., values={\'Name\': \'Alonso Fernández de Avellaneda\'})  # values comes from the catalog definition\n\nbook.author.find_and_set({\'Name\': \'Miguel de Cervantes\'})  # may raise ValueError if no value found\nbook.save(\'changed an author to the real one\')\n```\n\n### Catalog Enum fields\n\nEnums can be mapped to catalog items by ID or by custom property name.\n\n#### Enums mapped to specific catalog items ID\n\nNo catalog lookups are preformed on reading or writing of such fields.\n\n```python\nclass Genre(Enum):\n    fiction = 100001\n    nonfiction = 100002\n\n\nclass Book(PyrusModel):\n    genre = CatalogEnumField(<field_id>, catalog_id=<catalog_id>, enum=Genre, id_field=\'item_id\')\n\nbook = Book.objects.get(id=...)\n\nbook.genre\n>>> Genre.fiction\n\nbook.genre = Genre.nonfiction\nbook.save()\n\nbook.genre\n>>> Genre.nonfiction\n```\n\n\n#### Enums mapped to catalog item properties\n\n(imagine book has a property \'media\' with field \'Name\')\n\n```python\nclass Media(Enum):\n    paper = \'paper\'\n    papirus = \'papirus\'\n    pdf = \'pdf\'\n\nclass Book(PyrusModel):\n    media = CatalogEnumField(<field_id>, catalog_id=<catalog_id>, enum=Genre, id_field=\'Name\')\n```\n\n### Filtering\n\nOnly basic filtering is supported:\n\n```python\n\nBook.objects.get_filtered(\n    title=\'Don Quixote\',\n)\n>>> [Book(...), ...]\n\n\nBook.objects.get_filtered(\n    genre=Book.genre.find({\'Name\': \'Fiction\'})\n)\n>>> [Book(...), ...]\n\nBook.objects.get_filtered(\n    ...\n    include_archived=True,\n    steps=[1, 2],\n)\n>>> [Book(...), ...]\n```\n\n\n### Catalog fields, all the API\n```python\n# Read values\n\n# Non-empty value\nbook.author\n>>> CatalogItem(item_id=..., values={<your custom values here>})\n\nassert bool(book.author) == True\n\n# Empty value\nbook.author\n>>> CatalogEmptyValue()\n\nassert bool(book.author) == False\n\n\n# Get all possible values (works for empty fields as well)\nbook.author.catalog()\n>>> [CatalogItem(...), CatalogItem(...), ...]\n\n\n# Find a value in a catalog\nnew_author = book.author.catalog().find({\'Name\': \'Miguel de Cervantes\'})\nnew_author\n>>> CatalogItem(item_id=..., values={\'Name\': \'Miguel de Cervantes\'})  # or None\n\nbook.author = new_author\nbook.save()\n\n\n# Find and set shortcut\nbook.author.catalog().find_and_set({\'Name\': \'William Shakespeare\'})\n\nbook.author.find_and_set({\'Name\': \'NonExistent\'})\n>>> ValueError raised\n\n\n# Set value to a specific item_id\nbook.author = CatalogItem(item_id=123456)\n```',
    'author': 'Alexey Sveshnikov',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Callable, Generic, TypeVar, Type

T = TypeVar("T")
R = TypeVar("R")


class classproperty(Generic[T, R]):
    def __init__(self, func: Callable[[Type[T]], R]) -> None:
        self.func = func

    def __get__(self, obj: Any, cls: Type[T]) -> R:
        return self.func(cls)
